@@ -1,16 +1,17 @@
-// Copyright 2012-2013 Tinyarray authors.
+// Copyright 2012-2016 Tinyarray authors.
 //
 // This file is part of Tinyarray.  It is subject to the license terms in the
-// LICENSE file found in the top-level directory of this distribution and at
-// http://git.kwant-project.org/tinyarray/about/LICENSE.  A list of Tinyarray
-// authors can be found in the README file at the top-level directory of this
-// distribution and at http://git.kwant-project.org/tinyarray/about/.
-
+// file LICENSE.rst found in the top-level directory of this distribution and
+// at https://gitlab.kwant-project.org/kwant/tinyarray/blob/master/LICENSE.rst.
+// A list of Tinyarray authors can be found in the README.rst file at the
+// top-level directory of this distribution and at
+// https://gitlab.kwant-project.org/kwant/tinyarray.
 
 #include <Python.h>
 #include <cstddef>
 #include <sstream>
 #include <limits>
+#include <algorithm>
 #include "array.hh"
 #include "arithmetic.hh"
 #include "functions.hh"
@@ -70,11 +71,6 @@ Dtype dtype_of_scalar(PyObject *obj)
     if (PyLong_Check(obj)) return LONG;
     if (PyObject_HasAttr(obj, index_str)) return LONG;
 
-    // I'm not sure about this paragraph.  Does the existence of a __complex__
-    // method already signify that the number is to be interpreted as a complex
-    // number?  What about __float__?  Perhaps the following code is useless.
-    // In practice (with built-in and numpy numerical types) it never plays a
-    // role anyway.
     if (PyObject_HasAttr(obj, complex_str)) return COMPLEX;
     if (PyObject_HasAttr(obj, float_str)) return DOUBLE;
     if (PyObject_HasAttr(obj, int_str)) return LONG;
@@ -86,7 +82,7 @@ Dtype dtype_of_scalar(PyObject *obj)
 
 Dtype dtype_of_buffer(Py_buffer *view)
 {
-    char *fmt = view->format;
+    const char *fmt = view->format;
     Dtype dtype = NONE;
 
     // Currently, we only understand native endianness and alignment.
@@ -170,22 +166,23 @@ const char *seq_err_msg =
     "A sequence does not support sequence protocol - "
     "this is probably due to a bug in numpy for 0-d arrays.";
 
-// This function determines the shape of an array like sequence (or sequence of
-// sequences or number) given to it as first parameter.  `dtype_guess' is the
-// dtype of the first element of the sequence.
+// This function determines the shape of an array-like sequence (of
+// sequences...) given to it as first parameter.  `dtype_guess' is the dtype of
+// the first element of the sequence.
 //
 // All four arguments after the first one are written to. `shape' and `seqs'
 // must have space for at least `max_ndim' elements.
 //
 // After successful execution `seqs' will contain `ndim' new references
 // returned by PySequence_Fast.
-int examine_arraylike(PyObject *arraylike, int *ndim, size_t *shape,
-                      PyObject **seqs, Dtype *dtype_guess)
+int examine_sequence(PyObject *arraylike, int *ndim, size_t *shape,
+                     PyObject **seqs, Dtype *dtype_guess)
 {
     PyObject *p = arraylike;
     int d = -1;
-    while (true) {
-        if (PySequence_Check(p)) {
+    assert(PySequence_Check(p));
+    for (bool is_sequence = true; ; is_sequence = PySequence_Check(p)) {
+        if (is_sequence) {
             ++d;
             if (d == ptrdiff_t(max_ndim)) {
                 // Strings are, in a way, infinitely nested sequences because
@@ -210,7 +207,7 @@ int examine_arraylike(PyObject *arraylike, int *ndim, size_t *shape,
             break;
         }
 
-        // See http://projects.scipy.org/numpy/ticket/2199.
+        // See https://github.com/numpy/numpy/issues/652.
         seqs[d] = PySequence_Fast(p, seq_err_msg);
         if (!seqs[d]) {--d; goto fail;}
 
@@ -230,22 +227,13 @@ fail:
     return -1;
 }
 
-// This function is designed to be run after examine_arraylike.  It takes care
+// This function is designed to be run after examine_sequence.  It takes care
 // of releasing the references passed to it in seqs.
 template <typename T>
-int readin_arraylike(T *dest, int ndim, const size_t *shape,
-                     PyObject *arraylike, PyObject **seqs, bool exact)
+int readin_seqs(PyObject **seqs, T *dest, int ndim, const size_t *shape,
+                bool exact)
 {
-    if (ndim == 0) {
-        T value;
-        if (exact)
-            value = number_from_pyobject_exact<T>(arraylike);
-        else
-            value = number_from_pyobject<T>(arraylike);
-        if (value == T(-1) && PyErr_Occurred()) return -1;
-        *dest++ = value;
-        return 0;
-    }
+    assert(ndim > 0);
 
     // seqs is the stack of sequences being processed, all returned by
     // PySequence_Fast.  ps[d] and es[d] are the begin and end of the elements
@@ -301,7 +289,7 @@ int readin_arraylike(T *dest, int ndim, const size_t *shape,
             }
         }
 
-        // See http://projects.scipy.org/numpy/ticket/2199.
+        // See https://github.com/numpy/numpy/issues/652.
         seqs[d] = PySequence_Fast(*ps[d]++, seq_err_msg);
         if (!seqs[d]) {--d; goto fail;}
         len = PySequence_Fast_GET_SIZE(seqs[d]);
@@ -328,9 +316,8 @@ fail:
 }
 
 template <typename T>
-PyObject *make_and_readin_array(PyObject *in, int ndim_in, int ndim_out,
-                                const size_t *shape_out, PyObject **seqs,
-                                bool exact)
+PyObject *readin_seqs_into_new(PyObject **seqs, int ndim_in, int ndim_out,
+                               const size_t *shape_out, bool exact)
 {
     Array<T> *result = Array<T>::make(ndim_out, shape_out);
     assert(ndim_out >= ndim_in);
@@ -339,8 +326,8 @@ PyObject *make_and_readin_array(PyObject *in, int ndim_in, int ndim_out,
         assert(shape_out[d] == 1);
 #endif
     if (result == 0) return 0;
-    if (readin_arraylike<T>(result->data(), ndim_in,
-                            shape_out + ndim_out - ndim_in, in, seqs, exact)
+    if (readin_seqs<T>(seqs, result->data(), ndim_in,
+                       shape_out + ndim_out - ndim_in, exact)
         == -1) {
         Py_DECREF(result);
         return 0;
@@ -348,9 +335,33 @@ PyObject *make_and_readin_array(PyObject *in, int ndim_in, int ndim_out,
     return (PyObject*)result;
 }
 
-PyObject *(*make_and_readin_array_dtable[])(
-    PyObject*, int, int, const size_t*, PyObject**, bool) =
-    DTYPE_DISPATCH(make_and_readin_array);
+PyObject *(*readin_seqs_into_new_dtable[])(
+    PyObject**, int, int, const size_t*, bool) =
+    DTYPE_DISPATCH(readin_seqs_into_new);
+
+template <typename T>
+PyObject *readin_scalar_into_new(PyObject *in, bool exact, int ndim = 0)
+{
+    T value;
+    if (exact)
+        value = number_from_pyobject_exact<T>(in);
+    else
+        value = number_from_pyobject<T>(in);
+
+    if (value == T(-1) && PyErr_Occurred()) return 0;
+
+    Array<T> *result = Array<T>::make(ndim, 1);
+    *result->data() = value;
+
+    size_t *shape;
+    result->ndim_shape(0, &shape);
+    for (int d = 0; d < ndim; ++d) shape[d] = 1;
+
+    return (PyObject*)result;
+}
+
+PyObject *(*readin_scalar_into_new_dtable[])(PyObject*, bool, int) =
+    DTYPE_DISPATCH(readin_scalar_into_new);
 
 int examine_buffer(PyObject *in, Py_buffer *view, Dtype *dtype)
 {
@@ -463,6 +474,8 @@ T (*get_buffer_converter(Py_buffer *view))(const void *)
 template<typename T>
 int readin_buffer(T *dest, Py_buffer *view)
 {
+    if (view->len == 0) return 0;
+
     T (*number_from_ptr)(const void *) = get_buffer_converter<T>(view);
     if (!number_from_ptr) return -1;
 
@@ -494,6 +507,7 @@ int readin_buffer(T *dest, Py_buffer *view)
 
             for (int i = view->ndim - 1; i > 0; i--) {
                 if (indices[i] < view->shape[i]) break;
+                assert(indices[i] == view->shape[i]);
                 indices[i-1]++;
                 indices[i] = 0;
             }
@@ -505,11 +519,12 @@ int readin_buffer(T *dest, Py_buffer *view)
             *dest++ = (*number_from_ptr)(ptr);
             if (PyErr_Occurred()) return -1;
 
-            indices[view->ndim-1] ++;
+            indices[view->ndim-1]++;
             ptr += view->strides[view->ndim-1];
 
             for (int i = view->ndim - 1; i > 0; i--) {
                 if (indices[i] < view->shape[i]) break;
+                assert(indices[i] == view->shape[i]);
                 indices[i-1]++;
                 ptr += view->strides[i-1];
                 indices[i] = 0;
@@ -791,56 +806,76 @@ fail:
 // in Python.  As tinyarrays compare equal to equivalent tuples it is important
 // for the hashes to agree.  If not, there will be problems with dictionaries.
 
-long hash(long x)
+long old_hash(long x)
 {
     return x != -1 ? x : -2;
 }
 
-long hash(double x)
+#if PY_MAJOR_VERSION >= 3
+
+// the only documentation for this is in the Python sourcecode
+typedef Py_hash_t Hash;
+const Hash HASH_IMAG = _PyHASH_IMAG;
+
+Hash hash(long x_)
 {
-    const double two_to_31st = 2147483648.0;
-    double intpart, fractpart;
+    int negative = x_ < 0;
+    unsigned long x = (negative ? -x_ : x_);
+    // x is the absolute value of x_.
 
-    if (x == std::numeric_limits<double>::infinity())
-        return 314159;
-    else if (x == -std::numeric_limits<double>::infinity())
-        return -271828;
-    else if (x != x)
-        return 0;               // NaN
-
-    fractpart = modf(x, &intpart);
-    if (fractpart == 0) {
-        // This must return the same hash as an equal int or long.
-
-        if (intpart > std::numeric_limits<long>::max() ||
-            intpart < std::numeric_limits<long>::min()) {
-            // Convert to Python long and use its hash.
-            PyObject *plong = PyLong_FromDouble(x);
-            if (plong == NULL) return -1;
-            long result = PyObject_Hash(plong);
-            Py_DECREF(plong);
-            return result;
-        }
-        return hash(long(intpart));
+    if (x < PyLong_BASE) {
+        if (negative)
+            return x == 1 ? -2 : -long(x);
+        else
+            return x;
     }
 
-    int expo;
-    x = frexp(x, &expo) * two_to_31st;
-    long hipart = x;                        // Take the top 32 bits.
-    x = (x - double(hipart)) * two_to_31st; // Get the next 32 bits.
-    return hash(hipart + long(x) + (expo << 15));
+    // PyLong_SHIFT is 15 for 32-bit architectures and 30 for 64-bit.  So a
+    // long contains at most 3 digits.  Therefore, a starting value of n = 2 is
+    // sufficient.
+    Py_uhash_t r = 0;
+    for (int n = 2; n >= 0; --n) {
+        unsigned dig = (x >> (n * PyLong_SHIFT)) & PyLong_MASK;
+        if (r == 0 && dig == 0) continue;
+        r = ((r << PyLong_SHIFT) & _PyHASH_MODULUS) |
+            (r >> (_PyHASH_BITS - PyLong_SHIFT));
+        r += dig;
+        if (r >= _PyHASH_MODULUS) r -= _PyHASH_MODULUS;
+    }
+    if (negative) r = -r;
+    if (r == Py_uhash_t(-1)) r = Py_uhash_t(-2);
+    return r;
 }
 
-long hash(Complex x)
+#else
+
+typedef long Hash;
+const Hash HASH_IMAG = 1000003L;
+
+Hash hash(long x)
+{
+    return old_hash(x);
+}
+
+#endif
+
+Hash hash(double x)
+{
+    // We used to have our own implementation of this, but the extra function
+    // call is quite negligible compared to the execution time of the function.
+    return _Py_HashDouble(x);
+}
+
+Hash hash(Complex x)
 {
     // x.imag == 0  =>  hash(x.imag) == 0  =>  hash(x) == hash(x.real)
-    return hash(x.real()) + 1000003L * hash(x.imag());
+    return hash(x.real()) + HASH_IMAG * hash(x.imag());
 }
 
 // This routine calculates the hash of a multi-dimensional array.  The hash is
 // equal to that of an arrangement of nested tuples equivalent to the array.
 template <typename T>
-long hash(PyObject *obj)
+Hash hash(PyObject *obj)
 {
     int ndim;
     size_t *shape;
@@ -871,57 +906,145 @@ long hash(PyObject *obj)
                 r[d] = r_init;
             }
         } else {
-            if (d == 0) return hash(r[0] + r_addend);
+            if (d == 0) return old_hash(r[0] + r_addend);
             --d;
-            r[d] = (r[d] ^ hash(r[d+1] + r_addend)) * mult[d];
+            r[d] = (r[d] ^ old_hash(r[d+1] + r_addend)) * mult[d];
             mult[d] += mul_addend + 2 * i[d];
         }
     }
 }
 
 template <typename T>
-bool is_equal_data(PyObject *a_, PyObject *b_, size_t size)
+bool compare_scalar(const int op, const T a, const T b) {
+    switch(op){
+        case Py_EQ: return a == b;
+        case Py_NE: return a != b;
+        case Py_LE: return a <= b;
+        case Py_GE: return a >= b;
+        case Py_LT: return a < b;
+        case Py_GT: return a > b;
+        default:
+           assert(false);  // If we get here something is very wrong.
+           return false;   // Stop the compiler complaining.
+    }
+}
+
+
+template <>
+bool compare_scalar<Complex>(const int op, const Complex a, const Complex b) {
+    switch(op){
+        case Py_EQ: return a == b;
+        case Py_NE: return a != b;
+        // This function is never called in a context where
+        // the following code path is run -- fall through.
+        case Py_LE:
+        case Py_GT:
+        case Py_LT:
+        case Py_GE:
+        default:
+           assert(false);
+           return false;   // Stop the compiler complaining.
+    }
+}
+
+
+template <typename T>
+bool compare_data(int op, PyObject *a_, PyObject *b_, size_t size)
 {
     assert(Array<T>::check_exact(a_)); Array<T> *a = (Array<T>*)a_;
     assert(Array<T>::check_exact(b_)); Array<T> *b = (Array<T>*)b_;
-    T *data_a = a->data();
-    T *data_b = b->data();
-    for (size_t i = 0; i < size; ++i)
-        if (data_a[i] != data_b[i]) return false;
-    return true;
+    const T *data_a = a->data();
+    const T *data_b = b->data();
+    // Sequences are ordered the same as their first differing elements, see:
+    // https://docs.python.org/3/reference/expressions.html#value-comparisons
+    // comparison for "multidimensional" sequences is identical to comparing
+    // the flattened sequences when they have the same shape (the present case).
+    size_t i = 0;
+    for (; i < size; ++i)
+        if (data_a[i] != data_b[i]) break;
+    // Any of these operations should return true when objects are equal.
+    if (i == size) return ((op == Py_EQ) || (op == Py_LE) || (op == Py_GE));
+    // encapsulate this into a function to handle the COMPLEX case
+    return compare_scalar<T>(op, data_a[i], data_b[i]);
 }
 
-bool (*is_equal_data_dtable[])(PyObject*, PyObject*, size_t) =
-    DTYPE_DISPATCH(is_equal_data);
+bool (*compare_data_dtable[])(int, PyObject*, PyObject*, size_t) =
+    DTYPE_DISPATCH(compare_data);
+
+#if PY_MAJOR_VERSION < 3
+    #define PY2_RAISE_TYPEERROR(msg)\
+        PyErr_SetString(PyExc_TypeError, msg);\
+        return 0
+#else
+    #define PY2_RAISE_TYPEERROR(msg)
+#endif  // PY_MAJOR_VERSION < 3
 
 PyObject *richcompare(PyObject *a, PyObject *b, int op)
 {
-    if (op != Py_EQ && op != Py_NE) {
-        Py_INCREF(Py_NotImplemented);
-        return Py_NotImplemented;
-    }
+    PyObject *result;
+    const bool equality_comparison = (op == Py_EQ || op == Py_NE);
 
+    // Short circuit when we are comparing the same object.
     bool equal = (a == b);
-    if (equal) goto done;
+    if (equal) {
+        // Any of these operations should return true when objects are equal
+        equal = (op == Py_EQ) || (op == Py_GE) || (op == Py_LE);
+        result = equal ? Py_True : Py_False;
+        goto done;
+    }
 
     Dtype dtype;
     if (coerce_to_arrays(&a, &b, &dtype) < 0) return 0;
 
+    // Obviate the need for `compare_scalar<Complex` to
+    // handle the case of an undefined comparison.
+    if (dtype == COMPLEX && !equality_comparison) {
+        PY2_RAISE_TYPEERROR("unorderable type: complex()");
+        result = Py_NotImplemented;
+        goto decref_then_done;
+    }
+
     int ndim_a, ndim_b;
     size_t *shape_a, *shape_b;
     reinterpret_cast<Array_base*>(a)->ndim_shape(&ndim_a, &shape_a);
     reinterpret_cast<Array_base*>(b)->ndim_shape(&ndim_b, &shape_b);
-    if (ndim_a != ndim_b) goto decref_then_done;
-    for (int d = 0; d < ndim_a; ++d)
-        if (shape_a[d] != shape_b[d]) goto decref_then_done;
-    equal = is_equal_data_dtable[int(dtype)](a, b, calc_size(ndim_a, shape_a));
 
+    // TODO: Enable array comparisons between arrays of differing dimensions.
+    if (ndim_a != ndim_b) {
+        if (equality_comparison) {
+            goto equality_then_done;
+        } else {
+            PY2_RAISE_TYPEERROR("Unorderable type: only arrays with"
+                                "the same shape can be ordered.");
+            result = Py_NotImplemented;
+            goto decref_then_done;
+        }
+    }
+    for (int d = 0; d < ndim_a; ++d) {
+        if (shape_a[d] != shape_b[d]) {
+            if (equality_comparison) {
+                goto equality_then_done;
+            } else {
+                PY2_RAISE_TYPEERROR("Unorderable type: only arrays with"
+                                    "the same shape can be ordered.");
+                result = Py_NotImplemented;
+                goto decref_then_done;
+            }
+        }
+    }
+
+    // Actually compare the data.
+    equal = compare_data_dtable[int(dtype)](op, a, b, calc_size(ndim_a, shape_a));
+    result = equal ? Py_True : Py_False;
+    goto decref_then_done;
+
+// Non error-path exit points from this function
+equality_then_done:
+    result = ((op == Py_EQ) == equal) ? Py_True : Py_False;
 decref_then_done:
     Py_DECREF(a);
     Py_DECREF(b);
-
 done:
-    PyObject *result = ((op == Py_EQ) == equal) ? Py_True : Py_False;
     Py_INCREF(result);
     return result;
 }
@@ -935,6 +1058,7 @@ PyObject *get_dtype_py(PyObject *self, void *)
     };
     int dtype = int(get_dtype(self));
     assert(dtype < int(NONE));
+    Py_INCREF(dtypes[dtype]);
     return dtypes[dtype];
 }
 
@@ -1128,9 +1252,9 @@ template PyObject *str<long>(PyObject*);
 template PyObject *str<double>(PyObject*);
 template PyObject *str<Complex>(PyObject*);
 
-template long hash<long>(PyObject*);
-template long hash<double>(PyObject*);
-template long hash<Complex>(PyObject*);
+template Hash hash<long>(PyObject*);
+template Hash hash<double>(PyObject*);
+template Hash hash<Complex>(PyObject*);
 
 template int getbuffer<long>(PyObject*, Py_buffer*, int);
 template int getbuffer<double>(PyObject*, Py_buffer*, int);
@@ -1186,7 +1310,7 @@ Whenever an operation is missing from Tinyarray, NumPy can be used directly,\n\
 e.g.: numpy.linalg.det(my_tinyarray).");
 
 extern "C"
-void inittinyarray()
+MOD_INIT_FUNC(tinyarray)
 {
     // Determine storage formats.
     bool be = is_big_endian();
@@ -1205,11 +1329,12 @@ void inittinyarray()
     else
         format_by_dtype[int(LONG)] = UNKNOWN;
 
-    if (PyType_Ready(&Array<long>::pytype) < 0) return;
-    if (PyType_Ready(&Array<double>::pytype) < 0) return;
-    if (PyType_Ready(&Array<Complex>::pytype) < 0) return;
+    if (PyType_Ready(&Array<long>::pytype) < 0) return MOD_ERROR_VAL;
+    if (PyType_Ready(&Array<double>::pytype) < 0) return MOD_ERROR_VAL;
+    if (PyType_Ready(&Array<Complex>::pytype) < 0) return MOD_ERROR_VAL;
 
-    PyObject *m = Py_InitModule3("tinyarray", functions, tinyarray_doc);
+    PyObject *m;
+    MOD_DEF(m, "tinyarray", functions, tinyarray_doc);
 
     reconstruct = PyObject_GetAttrString(m, "_reconstruct");
 
@@ -1239,15 +1364,17 @@ void inittinyarray()
     // interpreter does the same, see try_complex_special_method in
     // complexobject.c
     int_str = PyString_InternFromString("__int__");
-    if (int_str == 0) return;
+    if (int_str == 0) return MOD_ERROR_VAL;
     long_str = PyString_InternFromString("__long__");
-    if (long_str == 0) return;
+    if (long_str == 0) return MOD_ERROR_VAL;
     float_str = PyString_InternFromString("__float__");
-    if (float_str == 0) return;
+    if (float_str == 0) return MOD_ERROR_VAL;
     complex_str = PyString_InternFromString("__complex__");
-    if (complex_str == 0) return;
+    if (complex_str == 0) return MOD_ERROR_VAL;
     index_str = PyString_InternFromString("__index__");
-    if (complex_str == 0) return;
+    if (complex_str == 0) return MOD_ERROR_VAL;
+
+    return MOD_SUCCESS_VAL(m);
 }
 
 int load_index_seq_as_long(PyObject *obj, long *out, int maxlen)
@@ -1255,6 +1382,7 @@ int load_index_seq_as_long(PyObject *obj, long *out, int maxlen)
     assert(maxlen >= 1);
     int len;
     if (PySequence_Check(obj)) {
+        // get a new reference -- don't forget to DECREF on all codepaths
         obj = PySequence_Fast(obj, "Bug in tinyarray, load_index_seq_as_long");
         if (!obj) return -1;
         Py_ssize_t long_len = PySequence_Fast_GET_SIZE(obj);
@@ -1272,6 +1400,7 @@ int load_index_seq_as_long(PyObject *obj, long *out, int maxlen)
             Py_DECREF(index);
             if (*out == -1 && PyErr_Occurred()) goto fail;
         }
+        Py_DECREF(obj);
     } else {
         len = 1;
         *out = PyInt_AsLong(obj);
@@ -1303,7 +1432,8 @@ int load_index_seq_as_ulong(PyObject *obj, unsigned long *uout,
 // If *dtype == NONE the simplest fitting dtype (at least dtype_min)
 // will be used and written back to *dtype.  Any other value of *dtype requests
 // an array of the given dtype.
-PyObject *array_from_arraylike(PyObject *in, Dtype *dtype, Dtype dtype_min)
+PyObject *array_from_arraylike(PyObject *in, Dtype *dtype, Dtype dtype_min,
+                               bool as_matrix)
 {
     Dtype dtype_in = get_dtype(in), dt = *dtype;
     int ndim;
@@ -1313,113 +1443,44 @@ PyObject *array_from_arraylike(PyObject *in, Dtype *dtype, Dtype dtype_min)
         // `in` is already an array.
         if (dt == NONE)
             dt = Dtype(std::max(int(dtype_in), int(dtype_min)));
-        if (dt == dtype_in)
-            Py_INCREF(result = in);
-        else
-            result = convert_array(dt, in, dtype_in);
+        if (as_matrix) {
+            size_t *in_shape;
+            reinterpret_cast<Array_base*>(in)->ndim_shape(&ndim, &in_shape);
+            if (ndim == 2) {
+                if (dt == dtype_in)
+                    Py_INCREF(result = in);
+                else
+                    result = convert_array(dt, in, dtype_in);
+            } else if (ndim < 2) {
+                shape[1] = (ndim == 0) ? 1 : in_shape[0];
+                shape[0] = 1;
+                result = convert_array(dt, in, dtype_in, 2, shape);
+            } else {
+                PyErr_SetString(PyExc_ValueError,
+                                "Matrix must be 2-dimensional.");
+                result = 0;
+            }
+        } else {
+            if (dt == dtype_in)
+                Py_INCREF(result = in);
+            else
+                result = convert_array(dt, in, dtype_in);
+        }
 
         *dtype = dt;
         return result;
-    } else {
-        // `in` is not an array.
+    } else if(PySequence_Check(in)) {
+        // `in` is not an array, but is a sequence.
 
         bool find_type = (dt == NONE);
 
-        // Try if buffer interface is supported
+        // Try if buffer interface is supported.
         Py_buffer view;
         if (examine_buffer(in, &view, find_type ? &dt : 0) == 0) {
             if (find_type && int(dt) < int(dtype_min)) dt = dtype_min;
             for (int i = 0; i < view.ndim; i++)
                 shape[i] = view.shape[i];
-            result = make_and_readin_buffer_dtable[int(dt)](&view, view.ndim,
-                                                            shape);
-            PyBuffer_Release(&view);
-
-            *dtype = dt;
-            return result;
-        }
-
-        if (examine_arraylike(in, &ndim, shape, seqs,
-                              find_type ? &dt : 0) == 0) {
-            if (find_type) {
-                PyObject *seqs_copy[max_ndim];
-                for (int d = 0; d < ndim; ++d)
-                    Py_INCREF(seqs_copy[d] = seqs[d]);
-                if (dt == NONE) {
-                    assert(shape[ndim - 1] == 0);
-                    dt = default_dtype;
-                }
-                if (int(dt) < int(dtype_min)) dt = dtype_min;
-                while (true) {
-                    result = make_and_readin_array_dtable[int(dt)](
-                        in, ndim, ndim, shape, seqs, true);
-                    if (result) break;
-                    dt = Dtype(int(dt) + 1);
-                    if (dt == NONE) {
-                        result = 0;
-                        break;
-                    }
-                    PyErr_Clear();
-                    for (int d = 0; d < ndim; ++d)
-                        Py_INCREF(seqs[d] = seqs_copy[d]);
-                }
-                for (int d = 0; d < ndim; ++d) Py_DECREF(seqs_copy[d]);
-            } else {
-                // A specific dtype has been requested.
-                result = make_and_readin_array_dtable[int(dt)](
-                    in, ndim, ndim, shape, seqs, false);
-            }
-
-            *dtype = dt;
-            return result;
-        }
-    }
-
-    return 0;
-}
-
-// If *dtype == NONE the simplest fitting dtype (at least dtype_min)
-// will be used and written back to *dtype.  Any other value of *dtype requests
-// an array of the given dtype.
-PyObject *matrix_from_arraylike(PyObject *in, Dtype *dtype, Dtype dtype_min)
-{
-    Dtype dtype_in = get_dtype(in), dt = *dtype;
-    int ndim;
-    size_t shape[max_ndim];
-    PyObject *seqs[max_ndim], *result;
-    if (dtype_in != NONE) {
-        // `in` is already an array.
-        if (dt == NONE)
-            dt = Dtype(std::max(int(dtype_in), int(dtype_min)));
-        size_t *in_shape;
-        reinterpret_cast<Array_base*>(in)->ndim_shape(&ndim, &in_shape);
-        if (ndim == 2) {
-            if (dt == dtype_in)
-                Py_INCREF(result = in);
-            else
-                result = convert_array(dt, in, dtype_in);
-        } else if (ndim < 2) {
-            shape[0] = 1;
-            shape[1] = (ndim == 0) ? 1 : in_shape[0];
-            result = convert_array(dt, in, dtype_in, 2, shape);
-        } else {
-            PyErr_SetString(PyExc_ValueError, "Matrix must be 2-dimensional.");
-            result = 0;
-        }
-
-        *dtype = dt;
-        return result;
-    } else {
-        // `in` is not an array.
-
-        bool find_type = (*dtype == NONE);
-
-        // Try if buffer interface is supported
-        Py_buffer view;
-        if (examine_buffer(in, &view, find_type ? &dt : 0) == 0) {
-            for (int i = 0; i < view.ndim; i++)
-                shape[i] = view.shape[i];
-            if (view.ndim != 2) {
+            if (as_matrix && view.ndim != 2) {
                 if (view.ndim > 2) {
                     PyErr_SetString(PyExc_ValueError,
                                     "Matrix must be 2-dimensional.");
@@ -1428,19 +1489,19 @@ PyObject *matrix_from_arraylike(PyObject *in, Dtype *dtype, Dtype dtype_min)
                 shape[1] = (view.ndim == 0) ? 1 : shape[0];
                 shape[0] = 1;
             }
-            if (find_type && int(dt) < int(dtype_min)) dt = dtype_min;
-            result = make_and_readin_buffer_dtable[int(dt)](&view, view.ndim,
-                                                            shape);
+            result = make_and_readin_buffer_dtable[int(dt)](
+                &view, (as_matrix ? 2 : view.ndim), shape);
             PyBuffer_Release(&view);
 
             *dtype = dt;
             return result;
         }
 
-        if (examine_arraylike(in, &ndim, shape, seqs,
-                              find_type ? &dt : 0) == 0) {
-            if (ndim != 2) {
+        if (examine_sequence(in, &ndim, shape, seqs,
+                             find_type ? &dt : 0) == 0) {
+            if (as_matrix && ndim != 2) {
                 if (ndim > 2) {
+                    for (int d = 0; d < ndim; ++d) Py_DECREF(seqs[d]);
                     PyErr_SetString(PyExc_ValueError,
                                     "Matrix must be 2-dimensional.");
                     return 0;
@@ -1449,19 +1510,17 @@ PyObject *matrix_from_arraylike(PyObject *in, Dtype *dtype, Dtype dtype_min)
                 shape[0] = 1;
             }
             if (find_type) {
-                // No specific dtype has been requested.  It will be
-                // determined by the input.
                 PyObject *seqs_copy[max_ndim];
                 for (int d = 0; d < ndim; ++d)
                     Py_INCREF(seqs_copy[d] = seqs[d]);
                 if (dt == NONE) {
-                    assert(shape[1] == 0);
+                    assert(shape[(as_matrix ? 2 : ndim) - 1] == 0);
                     dt = default_dtype;
                 }
                 if (int(dt) < int(dtype_min)) dt = dtype_min;
                 while (true) {
-                    result = make_and_readin_array_dtable[int(dt)](
-                        in, ndim, 2, shape, seqs, true);
+                    result = readin_seqs_into_new_dtable[int(dt)](
+                        seqs, ndim, (as_matrix ? 2 : ndim), shape, true);
                     if (result) break;
                     dt = Dtype(int(dt) + 1);
                     if (dt == NONE) {
@@ -1475,13 +1534,36 @@ PyObject *matrix_from_arraylike(PyObject *in, Dtype *dtype, Dtype dtype_min)
                 for (int d = 0; d < ndim; ++d) Py_DECREF(seqs_copy[d]);
             } else {
                 // A specific dtype has been requested.
-                result = make_and_readin_array_dtable[int(dt)](
-                    in, ndim, 2, shape, seqs, false);
+                result = readin_seqs_into_new_dtable[int(dt)](
+                    seqs, ndim, (as_matrix ? 2 : ndim), shape, false);
             }
 
             *dtype = dt;
             return result;
         }
+    } else {
+        // `in` is a scalar, or an invalid input.
+
+        dtype_in = dtype_of_scalar(in);
+        bool find_type = (dt == NONE);
+
+        if (dtype_in == NONE) {
+            PyErr_SetString(PyExc_TypeError,
+                            "Expecting an arraylike object or a scalar.");
+            return 0;
+        }
+
+        if (find_type) {
+            dt = Dtype(std::max(int(dtype_in), int(dtype_min)));
+            result = readin_scalar_into_new_dtable[int(dt)](
+                in, true, (as_matrix ? 2 : 0));
+        } else {
+            result = readin_scalar_into_new_dtable[int(dt)](
+                in, false, (as_matrix ? 2 : 0));
+        }
+
+        *dtype = dt;
+        return result;
     }
 
     return 0;
@@ -1577,6 +1659,13 @@ PyObject *conjugate(PyObject *in_, PyObject *)
   return apply_unary_ufunc<Conjugate<T> >(in_);
 }
 
+template <typename T>
+PyObject *size_of(PyObject *in_, PyObject *)
+{
+    assert(Array<T>::check_exact(in_));
+    return PyInt_FromSsize_t(((const Array<T>*) in_)->object_size());
+}
+
 template <typename T>
 Array<T> *Array<T>::make(int ndim, size_t size)
 {
@@ -1620,6 +1709,7 @@ Array<T> *Array<T>::make(int ndim, const size_t *shape, size_t *sizep)
     return result;
 }
 
+
 template <typename T>
 PyObject *reduce(PyObject *self_, PyObject*)
 {
@@ -1632,16 +1722,22 @@ PyObject *reduce(PyObject *self_, PyObject*)
     self->ndim_shape(&ndim, &shape);
     size_t size_in_bytes = calc_size(ndim, shape) * sizeof(T);
 
-    Py_INCREF(reconstruct);
     PyObject *pyshape = PyTuple_New(ndim);
     for (int i=0; i < ndim; ++i)
         PyTuple_SET_ITEM(pyshape, i, PyInt_FromSize_t(shape[i]));
     PyObject *format = PyInt_FromLong(format_by_dtype[int(get_dtype(self_))]);
-    PyObject *data = PyString_FromStringAndSize((char*)self->data(),
+    PyObject *data = PyBytes_FromStringAndSize((char*)self->data(),
                                                 size_in_bytes);
 
+    // PyTuple_SET_ITEM steals references, so we need to INCREF
+    Py_INCREF(reconstruct);
     PyTuple_SET_ITEM(result, 0, reconstruct);
+    // Py_BuildValue does not steal references, so we need to DECREF
     PyTuple_SET_ITEM(result, 1, Py_BuildValue("(OOO)", pyshape, format, data));
+    Py_DECREF(pyshape);
+    Py_DECREF(format);
+    Py_DECREF(data);
+
     return result;
 }
 
@@ -1668,6 +1764,13 @@ PyMappingMethods Array<T>::as_mapping = {
     getitem<T>      // mp_subscript
 };
 
+#if PY_MAJOR_VERSION >= 3
+template <typename T>
+PyBufferProcs Array<T>::as_buffer = {
+    getbuffer<T>,  // bf_getbuffer
+    0,             // bf_releasebuffer
+};
+#else
 template <typename T>
 PyBufferProcs Array<T>::as_buffer = {
     // We only support the new buffer protocol.
@@ -1677,15 +1780,25 @@ PyBufferProcs Array<T>::as_buffer = {
     0,                          // bf_getcharbuffer
     getbuffer<T> // bf_getbuffer
 };
+#endif
 
 template <typename T>
 PyMethodDef Array<T>::methods[] = {
     {"transpose", (PyCFunction)transpose<T>, METH_NOARGS},
     {"conjugate", (PyCFunction)conjugate<T>, METH_NOARGS},
+    {"__sizeof__", (PyCFunction)size_of<T>, METH_NOARGS},
     {"__reduce__", (PyCFunction)reduce<T>, METH_NOARGS},
     {0, 0}                      // Sentinel
 };
 
+#if PY_MAJOR_VERSION >= 3  // don't need flags for buffers or checking types
+    static unsigned long _tp_flags = Py_TPFLAGS_DEFAULT;
+#else
+    static long _tp_flags = Py_TPFLAGS_DEFAULT |
+        Py_TPFLAGS_HAVE_NEWBUFFER |
+        Py_TPFLAGS_CHECKTYPES;
+#endif
+
 template <typename T>
 PyTypeObject Array<T>::pytype = {
     PyVarObject_HEAD_INIT(&PyType_Type, 0)
@@ -1696,7 +1809,7 @@ PyTypeObject Array<T>::pytype = {
     0,                              // tp_print
     0,                              // tp_getattr
     0,                              // tp_setattr
-    0,                              // tp_compare
+    0,                              // tp_compare  (tp_reserved in Python 3.x)
     repr<T>,                        // tp_repr
     &as_number,                     // tp_as_number
     &as_sequence,                   // tp_as_sequence
@@ -1707,9 +1820,7 @@ PyTypeObject Array<T>::pytype = {
     PyObject_GenericGetAttr,        // tp_getattro
     0,                              // tp_setattro
     &as_buffer,                     // tp_as_buffer
-    Py_TPFLAGS_DEFAULT |
-    Py_TPFLAGS_HAVE_NEWBUFFER |
-    Py_TPFLAGS_CHECKTYPES,          // tp_flags
+    _tp_flags,                      // tp_flags
     0,                              // tp_doc
     0,                              // tp_traverse
     0,                              // tp_clear
@@ -1744,6 +1855,10 @@ template PyObject *conjugate<long>(PyObject*, PyObject*);
 template PyObject *conjugate<double>(PyObject*, PyObject*);
 template PyObject *conjugate<Complex>(PyObject*, PyObject*);
 
+template PyObject *size_of<long>(PyObject*, PyObject*);
+template PyObject *size_of<double>(PyObject*, PyObject*);
+template PyObject *size_of<Complex>(PyObject*, PyObject*);
+
 template PyObject *reduce<long>(PyObject*, PyObject*);
 template PyObject *reduce<double>(PyObject*, PyObject*);
 template PyObject *reduce<Complex>(PyObject*, PyObject*);
